#
# Class for contents of Gaussian matrix element file.
#
# Methods:
#
# new              -- create empty object.
#
# read(filename)   -- open file and read contents.
#
# write(filename)  -- open file and read contents.
#
# name             -- return element name (scalar or array ref).
# name(value)      -- set element name (scalar or array ref).
#
# item(name)
# item(name,value) -- set/return items by name.  case-insensitive
#                     and works for names with embedded spaces.
#
# add(name,objref) -- add a named object (typically an OpMat).
#
# Names of objects which are or may be present and are likely to
# be of general interest are listed below.  Name is
# case-insensitive.
#
# header(@items)   -- return an array with all items in the standard
#                     header records, with refs for the arrays, or
#                     set the items if arguments are provided.  The
#                     items are those in the exported arrays
#                     @head_scalars followed by those in @head_arrays.
#
# copy_header(oldfile) -- copy basic initial record information from
#                         another object to the current one.
#
# addopmat(name,type,dimsref,arrayref) -- create a new OpMat object and add.
#
# prtcon -- print out connectivity.
#
# Importable function:
# fixstr(len,string) -- return a fixed-length version of string, truncated or
#                       padded with blanks.
#
# Importable variables:
#
# @head_scalars -- names of the scalar values in the standard header records,
#                  in the order returned from/passed to header().
#
# @head_arrays  -- names of the array refs in the standard header records,
#                  in the order returned from/passed to header().
#
# %scalars      -- keys are the names of the floating scalar result values
#                  collected together (corresponding to Gaussian /Gen/).
#
# %stdmat       -- keys are the names of standard operator matrices which
#                  may be in a file.  When writing a file, these are
#                  written before any other operators, although the logic
#                  for processing the files does not depend on this.
#
# Scalars which are always present:
#
# ENucRep   -- nuclear repulsion energy
# JobStatus -- -1 in progress, 0 undefined, 1 finished successfully,
#              2 step in multi-step job completed successfully.
#              3 error termination.
#
# Other scalars are listed in the initialization for @head_scalars
# and documented in unfdat.txt
#
# Arrays which are always present:
# These are listed below in the initialization for @head_arrays and
# documented in unfdat.txt
#
# Scalars which may be present:
#
# Derived arrays which may be preent:
# $x->connectivity -- connectivity.  An array of lists, one per atom.
#                     Each list consists of zero or more (atnum,bondtype)
#                     pairs, giving an atom bonded to the owner of the
#                     list and the bond order.
#
# Arrays which may be present:
# These are listed in the unfdat.txt file which documents the data file
# format.
#
package MatEl;
use strict;
use warnings;
use IO::File;
use Carp;
use GauComplex;
use OpMat qw(&fixstr);
use Exporter;

our ($AUTOLOAD,@EXPORT_OK,@ISA);

@ISA = qw(Exporter);

@EXPORT_OK = qw(&fixstr %scalars %stdmat @head_scalars @head_arrays printcon @mat_names &aoint);

#
# Names in the hash start with one _ for operator objects (class OpMat) and
# with __ for internal/derived values/arrays (which are stored in the fixed
# records at the beginning of the file).
#
# names for the scalars in the /Gen/ array.
my (@scalar_names) = ("VIRIAL","X-EFIELD","Y-EFIELD","Z-EFIELD","TE SCF ENERGY","SCRF G-FACTOR","SCRF A0",
  "THERMAL ENERGY","ECC","ECC(T)","EVAR1","ZPE","COMPOUND ENERGY","NIMAG","DEPUHF","EPUHF","ECBS2","ECBSI",
  "EPMP2","EPMP3","GEOM RMSF","ECIS-MP2","SCF RMSDP","S2-A","ECIS","DEUMP4D","EBDREF","EMP5","S4SD","EFC",
  "SCFTAU","ESCF","EUMP2","EUMP3","EUMP4","CBS OIII","EPRF","EMP4DQ","EMP4SDQ","SCALAR40","ENUCREP",
  "PSCFT","ETOTAL","S2SCF","S2-1","S2-D","A0","SCALAR48","TEMPERATURE","PRESSURE","FREQSCALE","INACTNUCREP",
  "DE2-SINGLES","DE2","RF NUC","RF ELEC","CURVATURE","IRC RC","EXTFLAG","ESCF1IT","JOB STATUS","NDERIV");

my ($GSNAME) = "GAUSSIAN SCALARS";
my ($GSNAMEX) = "__" . $GSNAME;

# names for standard order of matrices in file, for writing
our (@mat_names) =  ("NUCLEAR GRADIENT","NUCLEAR FORCE CONSTANTS","ELECTRIC DIPOLE MOMENT",
  "ELECTRIC DIPOLE POLARIZABILITY","ELECTRIC DIPOLE DERIVATIVES",
  "DIPOLE POLARIZABILITY DERIVATIVES","ELECTRIC DIPOLE HYPERPOLARIZABILITY",
  "OVERLAP","CORE HAMILTONIAN ALPHA","CORE HAMILTONIAN BETA","KINETIC ENERGY",
  "ORTHOGONAL BASIS","DIPOLE INTEGRALS","QUADRUPOLE INTEGRALS",
  "OCTOPOLE INTEGRALS","HEXADECAPOLE INTEGRALS","DIP VEL INTEGRALS","R X DEL INTEGRALS",
  "ALPHA ORBITAL ENERGIES","BETA ORBITAL ENERGIES","ALPHA MO COEFFICIENTS",
  "BETA MO COEFFICIENTS","ALPHA DENSITY MATRIX","BETA DENSITY MATRIX",
  "ALPHA SCF DENSITY MATRIX","BETA SCF DENSITY MATRIX","ALPHA FOCK MATRIX",
  "BETA FOCK MATRIX","OVERLAP DERIVATIVES","CORE HAMILTONIAN DERIVATIVES",
  "F(X)","DENSITY DERIVATIVES","FOCK DERIVATIVES","ALPHA UX","BETA UX",
  "ALPHA MO DERIVATIVES","BETA MO DERIVATIVES","ALPHA SCF DENSITY","BETA SCF DENSITY",
  "ALPHA MP FIRST ORDER DENSITY","BETA MP FIRST ORDER DENSITY",
  "ALPHA MP2 DENSITY","BETA MP2 DENSITY","ALPHA MP3 DENSITY","BETA MP3 DENSITY",
  "ALPHA MP4 DENSITY","BETA MP4 DENSITY","ALPHA CI ONE-PARTICLE DENSITY",
  "BETA CI ONE-PARTICLE DENSITY","ALPHA CI DENSITY","BETA CI DENSITY",
  "ALPHA QCI/CC DENSITY","BETA QCI/CC DENSITY","ALPHA DENSITY CORRECT TO SECOND ORDER",
  "BETA DENSITY CORRECT TO SECOND ORDER","ALPHA ONIOM DENSITY","BETA ONIOM DENSITY",
  "GIAO D2H/DBDM","GIAO L/R3", "REGULAR 2E INTEGRALS", "RAFFENETTI 2E INTEGRALS",
  "AA MO 2E INTEGRALS", "AB MO 2E INTEGRALS", "BA MO 2E INTEGRALS", "BB MO 2E INTEGRALS");

our (@head_scalars) = ("title","natoms","nbasis","nbsuse","icharg","multip",
		       "ne","iopcl","icgu","nfc","nfv","itran","idum9",
		       "nshellao","nprimao","nshelldb","nprimdb","nbondtot");

our (@head_arrays) = ("ian","iattyp","atmchg","c","ibf","ibftyp","atmwgt");

my ($VERSION) = 1.0;

our (%scalars,%stdmat,%headscals);

BEGIN: {
  my ($i,$name);
  for ($i=0; $i < @scalar_names; $i++) {($name = $scalar_names[$i]) =~ s/ //g; $scalars{$name} = $i;}
  for (; $i < 1000 ; $i++) {$name = "SCALAR" . $i; $scalars{$name} = $i;}
# foreach $i (sort keys %scalars) {print "scalar key $i val $scalars{$i}\n";}
  for ($i=0; $i < @mat_names; $i++) {$stdmat{"_".$mat_names[$i]} = $i;}
  for ($i=0; $i < @head_scalars; $i++) {$headscals{"_".uc($head_scalars[$i])} = $i;}
# foreach $i (sort keys %headscals) {print "headscals key $i val $headscals{$i}\n";}
}

sub new {
  my $this = shift;
  my $class = ref($this) || $this;
  my $self = {};
  bless $self, $class;
  $self->_init(@_);
  return $self;
}

sub _init {
  my $self = shift;
  my ($debug,$key) = (0);
  if (@_) {$debug = $_[0];}
  foreach $key (keys %$self) {delete($$self{$key});}
  $$self{"__FH"} = undef;
  $$self{"__LABFIL"} = "Gaussian matrix elements                                        ";
  $$self{"__IVERS"} = 2;
  $$self{"__NLAB"} = 11;
  $$self{"__GVERS"} = fixstr(64,"MatEl $VERSION");
  $$self{"__LENREC"} = 4000;
  $$self{"__DEBUG"} = $debug;
  $$self{"__LEN12L"} = 4;
  $$self{"__LEN4L"} = 4;
  my (@rec11) = (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
  $$self{"__REC11"} = \@rec11;
}

sub read {
  my $self = shift;
  my $file = shift;
  $self->_open("<$file");
  my ($fh,$debug,$rec) = ($$self{"__FH"},$$self{"__DEBUG"});
  $self->_dorec1(0);
  $self->_dorec2(0);
  $$self{"__IAN"} = $self->_readvec("i",$$self{"__NATOMS"});
  $$self{"__IATTYP"} = $self->_readvec("i",$$self{"__NATOMS"});
  $$self{"__ATMCHG"} = $self->_readvec("d",$$self{"__NATOMS"});
  $$self{"__C"} = $self->_readvec("d",3*$$self{"__NATOMS"});
  my ($temp) = $self->_readvec("i",2*$$self{"__NBASIS"});
  my ($st,$end) = (0,$$self{"__NBASIS"}-1);
  my (@ibf) = @$temp[$st .. $end];
  $$self{"__IBF"} = \@ibf;
  ($st,$end) = ($$self{"__NBASIS"},2*$$self{"__NBASIS"}-1);
  my (@ibftyp) = @$temp[$st .. $end];
  $$self{"__IBFTYP"} = \@ibftyp;
  $$self{"__ATMWGT"} = $self->_readvec("d",$$self{"__NATOMS"});
  $temp = $self->_readvec("i",4);
  ($$self{"__NFC"},$$self{"__NFV"},$$self{"__ITRAN"},$$self{"__IDUM9"}) = @$temp;
  my ($ndo,$i,$namerec,$ix,$type) = $$self{"__NLAB"} - 10;
  my ($lenx) = $self->_readvec("i",$ndo);
  for ($i=0; $i < $ndo ; $i++) {
    $namerec = "__REC" . ($i+11);
    $$self{$namerec} = $self->_readvec("i",$$lenx[$i]);}
  if ($ndo > 0) {
    my $rec11 = $$self{"__REC11"};
    $self->{_intname("nshellao")} = $rec11->[0];
    $self->{_intname("nprimao")} = $rec11->[1];
    $self->{_intname("nshelldb")} = $rec11->[2];
    $self->{_intname("nprimdb")} = $rec11->[3];
    $self->{_intname("nbondtot")} = $rec11->[4];
  }
  while (1) {
    my (@labarr) = $self->_readlab();
    last if ($labarr[0] =~ /END/);
    my ($label,$ni,$nrx,$nval,$lenbuf,@dimens) = @labarr[0..9];
    my ($labobj,$oparr) = $label;
    $labobj =~ s/ *$//;
    ($type,$oparr) = $self->_readarr(@labarr);
    if ($labobj eq $GSNAME) {if ($debug > 1) {print " do scalars\n"}; $$self{$GSNAMEX} = $oparr;}
    else {
      if (aoint($label)) {$dimens[0]=$dimens[0];$dimens[1]=$dimens[0];$dimens[2]=$dimens[0];$dimens[4]=$nrx;}
      $self->addopmat ($label,$type,\@dimens,$oparr);}
  }
  if ($self->nbondtot > 0) {$self->_makecon();}
  $self->_close();
}

sub add {
  my $self = shift;
  my ($labobj,$obj) = @_;
  $labobj =~ s/ *$//;
  $labobj = "_" . uc($labobj);
  $$self{$labobj} = $obj;
}

sub addopmat {
  my $self = shift;
  my ($label,$type,$dims,$oparr) = @_;
  $label =~ s/ *$//;
  my($labobj) = "_" . uc($label);
  $$self{$labobj} = OpMat->new($label,$type,$dims,$oparr);
}

sub _open {
  my $self = shift;
  my ($file) = @_;
  $$self{"__FH"} = new IO::File;
  my ($fh,$debug) = ($$self{"__FH"},$$self{"__DEBUG"});
  if ($fh->open("$file")) {if ($debug) {print " opened $file\n";}}
  else {die "failed to open $file";}
}

sub _close {
  my $self = shift;
  if ($$self{"__FH"}) {$$self{"__FH"}->close() || die "failed to close";}
  $$self{"__FH"} = undef;
}

sub write {
  my $self = shift;
  my ($file) = @_;
  my ($debug) = $$self{"__DEBUG"};
  $self->_open(">$file");
  $$self{"__GVERS"} = fixstr(64,"MatEl $VERSION");
  if ($debug == 1) {print " write header records\n";}
  $self->_dorec1(1);
  $self->_dorec2(1);
  my ($nat,$nbf) = ($$self{"__NATOMS"},$$self{"__NBASIS"});
  $self->_rwvec(1,"i",$nat,$$self{"__IAN"});
  $self->_rwvec(1,"i",$nat,$$self{"__IATTYP"});
  $self->_rwvec(1,"d",$nat,$$self{"__ATMCHG"});
  $self->_rwvec(1,"d",3*$nat,$$self{"__C"});
  my ($len,$ribf,$ribftyp,@x) = (0,$$self{"__IBF"},$$self{"__IBFTYP"});
  splice(@x,@x,$nbf,@$ribf);
  splice(@x,@x,$nbf,@$ribftyp);
  $len = @x;
  $self->_rwvec(1,"i",$len,\@x);
  $self->_rwvec(1,"d",$nat,$$self{"__ATMWGT"});
  @x = ($$self{"__NFC"},$$self{"__NFV"},$$self{"__ITRAN"},$$self{__IDUM9});
  $self->_rwvec(1,"i",4,\@x);
  my ($ndo,$i,$key,$namerec,$ix,@lrecs,$ntot,$lenbuf) = $$self{"__NLAB"} - 10;
  if ($ndo >= 1) {
    my $rec11 = $$self{"__REC11"};
    $rec11->[0] = $self->nshellao;
    $rec11->[1] = $self->nprimao;
    $rec11->[2] = $self->nshelldb;
    $rec11->[3] = $self->nprimdb;}
  for ($i=0; $i < $ndo ; $i++) {
    $namerec = "__REC" . ($i+11);
    $ix = $$self{$namerec};
    $lrecs[$i] = @$ix;}
  $self->_rwvec(1,"i",$ndo,\@lrecs);
  for ($i=0; $i < $ndo ; $i++) {
    $namerec = "__REC" . ($i+11);
    $ix = $$self{$namerec};
    $self->_rwvec(1,"i",$lrecs[$i],$ix);}
  if (defined($$self{$GSNAMEX})) {
    $ix = $$self{$GSNAMEX};
    $ntot = @$ix;
    $lenbuf = $$self{"__LENREC"}/2;
    my($nval) = $self->_write_ir(0,1,$lenbuf,$ix);
    my (@dims) = ($ntot,1,1,1,1,0);
    $self->_writelab($GSNAME,1,1,$nval,$lenbuf,@dims);
    $self->_write_ir(1,1,$lenbuf,$ix);
    }
  for ($i = 0; $i < @mat_names; $i++) {
    $key = "_" . $mat_names[$i];
    if (defined($$self{$key})) {
      $self->_writeop($key);}}
  foreach $key (keys(%$self)) {
    next if ($key =~ /^__/);
    next if ($key !~ /^_/);
    next if (defined($stdmat{$key}));
    next if (defined($headscals{$key}));
    $ix = $$self{$key};
    if (ref($ix) eq "OpMat") {
      if ($debug > 1) {print " key $key val ",ref($ix)," OK\n";}
      $self->_writeop($key);}
    else {print " key $key val ",ref($ix)," BAD\n";}
  }
  $self->_writelab("END",0,0,0,0,0,0,0,0,0,0);
}

sub _writeop {
  my ($self) = shift;
  my ($key) = @_;
  my ($ix,$lenbuf,$oname,$otype,$odims,$oval,$ni,$nr,$nri,$i,$ntot);
  $lenbuf = $$self{"__LENREC"};
  $ix = $$self{$key};
  $oname = $ix->name();
  $otype = $ix->type();
  $odims = $ix->dimens();
  $oval = $ix->array();
  my (@dims) = (0,0,0,0,0,0);
  $ni = 0;
  if ($otype eq "c") {$ni = 0; $nr = -1; $nri = 2; $lenbuf = $lenbuf/4;}
  elsif ($otype eq "d") {$ni = 0; $nr = 1; $nri = 1; $lenbuf = $lenbuf/2;}
  else {$ni = 1; $nr = 0; $nri = 1;}
  $ntot = @$oval;
  for ($i=0; $i < @$odims; $i++) {$dims[$i] = $odims->[$i];}
  if (aoint($oname)) {
    $ni = 4;
    $nr = ($dims[4] > 0) ? $dims[4] : 1;
    $dims[0] = $$self{"__NBASIS"};
    $dims[1] = $dims[0];
    $dims[2] = $dims[0];
    $dims[3] = $dims[0];
    $dims[4] = 0;
    $ntot = $self->_write2e(0,$nr,$dims[0],$lenbuf,$oval);
    $self->_writelab($oname,$ni,$nr,$ntot,$lenbuf,@dims);
    $ntot = $self->_write2e(1,$nr,$dims[0],$lenbuf,$oval);}
  else {
    $self->_writelab($oname,$ni,$nr,$ntot,$lenbuf,@dims);
    $self->_writebufvec($nri,$otype,$ntot,$lenbuf,$oval);}
}

sub _readvec {
  my $self = shift;
  my (@x);
  $self->_rwvec(0,@_,\@x);
  return (\@x);
}

sub _rwvec {
  my $self = shift;
  my ($dowrt,$type,$n,$x) = @_;
  my ($debug) = $$self{"__DEBUG"};
  if ($debug > 1) {print " rwvec dowrt $dowrt type $type n $n\n";}
  my ($fh,$didrnd,$rec,$lrec,$pkstr) = ($$self{"__FH"},0);
  if ($type eq "i") {
    $lrec = $n * 4;
    if (($n % 2) == 1) {$didrnd = 1; $lrec += 4;}}
  elsif ($type eq "d") {$lrec = $n * 8;}
  else {die "bad type $type in MatEl::_readvec";}
  $pkstr = sprintf($type."%d",$n);
  if ($dowrt) {
    $rec = pack($pkstr,@$x);
    if ($didrnd) {$rec .= pack("i1",0);}
    $fh->write($rec,$lrec);}
  else {
    $fh->read($rec,$lrec);
    @$x = unpack($pkstr,$rec);}
}

sub _dorec1 {
  my $self = shift;
  my ($dowrt) = @_;
  my ($fh,$pkstr,$len,$rec) = ($$self{"__FH"},"a64i2a64",2*64+2*4);
  my ($debug) = $$self{"__DEBUG"};
  if ($debug > 1) {print " dorec1 dowrt $dowrt\n";}
  if ($dowrt) {
    my ($label) = fixstr(64,$$self{"__LABFIL"});
    my ($vers) = fixstr(64,$$self{"__GVERS"});
    $rec = pack($pkstr,$label,$$self{"__IVERS"},$$self{"__NLAB"},$vers);
    $fh->write($rec,$len);}
  else {
    $fh->read($rec,$len);
    ($$self{"__LABFIL"},$$self{"__IVERS"},$$self{"__NLAB"},$$self{"__GVERS"}) = unpack($pkstr,$rec);}
}

sub _dorec2 {
  my $self = shift;
  my ($dowrt) = @_;
  my ($fh,$pkstr,$len,$rec) = ($$self{"__FH"},"a64i10",64+10*4);
  my ($debug) = $$self{"__DEBUG"};
  if ($debug > 1) {print " dorec2 dowrt $dowrt\n";}
  if ($dowrt) {
    my ($title) = fixstr(64,$$self{"__TITLE"});
    $rec = pack($pkstr,$title,$$self{"__NATOMS"},$$self{"__NBASIS"},$$self{"__NBSUSE"},$$self{"__ICHARG"},$$self{"__MULTIP"},
		$$self{"__NE"},$$self{"__LEN12L"},$$self{"__LEN4L"},$$self{"__IOPCL"},$$self{"__ICGU"});
    $fh->write($rec,$len);
  }
  else {
    $fh->read($rec,$len);
    ($$self{"__TITLE"},$$self{"__NATOMS"},$$self{"__NBASIS"},$$self{"__NBSUSE"},$$self{"__ICHARG"},$$self{"__MULTIP"},
     $$self{"__NE"},$$self{"__LEN12L"},$$self{"__LEN4L"},$$self{"__IOPCL"},$$self{"__ICGU"}) = unpack($pkstr,$rec);}
}

sub _readlab {
  my $self = shift;
  my ($fh,$pkstr,$lrec,$rec,$nread) = ($$self{"__FH"},"a64i9",64+10*4);
  my ($label,@vals) = ("END",0,0,0,0,0,0,0,0,0);
  $nread = $fh->read($rec,$lrec);
  if ($nread == $lrec) {($label,@vals) = unpack($pkstr,$rec);}
  elsif ($nread != 0) {die "read error in MatEl::_readlab lrec $lrec nread $nread";}
  return ($label,@vals);
}

sub _writelab {
  my $self = shift;
  my ($label) = fixstr(64,$_[0]); shift;
  my ($fh,$debug,$pkstr,$lrec,$rec) = ($$self{"__FH"},$$self{"__DEBUG"},"a64i10",64+10*4);
  my (@vals) = @_;
  if ($debug) {print " write $label ",join(" ",@vals),"\n";}
  if (@vals != 10) {die "bad values ",join(" ",@vals)," in writelab\n";}
  my ($lx);
  $lx = @vals;
  $rec = pack($pkstr,$label,@vals);
  $fh->write($rec,$lrec);
}

sub _readarr {
  my $self = shift;
  my ($label,$ni,$nrx,$nval,$lenbuf,$n1,$n2,$n3,$n4,$n5) = @_;
  my ($nri,$nr,$type);
  if ($nrx > 0) {$nr = $nrx ; $nri = 1;}
  else {$nr = -$nrx; $nri = 2;}
  if (($nr > 0) && ($nri == 2)) {$type = "c";}
  elsif ($nr > 0) {$type = "d";}
  else {$type = "i";}
  my ($fh,$rec,$nread) = $$self{"__FH"};
  my ($n2345) = $n2 * $n3 * $n4 * $n5;
  my ($x);
# AO 2e integrals
  if (aoint($label)) {$x = $self->_read2e($label,$ni,$nr,$nval,$lenbuf,$n1);}
# 1d array packed with indices for non-zero values.
  elsif (($ni == 1) && ($nr == 1) && ($n2345 == 1)) {$x = $self->_read_ir($nri,$n1,$nval,$lenbuf);}
# unpacked integer array
  elsif (($ni > 0) && ($nr == 0)) {$x = $self->_readbufvec(1,"i",$nval,$lenbuf);}
# unpacked real arrays with various dimensions.
  elsif (($ni == 0) && ($nr == 1)) {$x = $self->_readbufvec($nri,"d",$nval,$lenbuf);}
  else {die "MatEl::_readarr is confused about this matrix @_";}
return($type,$x);
}

#
# Read a packed real array and unpack it.  The array
# is written as a sequence of index,value array pairs
# with length given by the buffer size parameter in
# the label record.  Arguments:
#
# $n      -- size of unpacked array.
# $nval   -- number of non-values
# $lenbuf -- length of each index and value array.
#
# returns a reference to the unpacked array.
#
sub _read_ir {
  my $self = shift;
  my ($nri,$n,$nval,$lenbuf) = @_;
  my ($fh,$ind,$buf,$i,$j,@x,$ndo,$indji) = $$self{"__FH"};
  if ($nri == 1) {$x[$n-1] = 0.0e0;}
  else {$x[$n-1] = cplx(0,0);}
  for ($i = 0; $i < ($n); $i++) {$x[$i]=$x[$n-1];}
  for ($i = 0; $i < $nval; $i += $lenbuf) {
    $ind = $self->_readvec("i",$lenbuf);
    $buf = $self->_readvec("d",$lenbuf);
    $ndo = $lenbuf;
    if (($i+$ndo) > $nval) {$ndo = $nval - $i;}
    if ($nri == 1) {
      for ($j = 0; $j < $ndo; $j++) {$x[$$ind[$j]-1] = $$buf[$j];}}
    else {
      my ($real,$imag);
      $indji = 0;
      for ($j = 0; $j < $ndo; $j++) {
	$real = $$buf[$indji++];
	$imag = $$buf[$indji++];
	$x[$$ind[$j]-1] = cplx($real,$imag);}}
  }
  return (\@x);
}

sub _write_ir {
  my $self = shift;
  my ($dowrt,$nri,$lenbuf,$ix) = @_;
  my ($nval,$indib,$indrb,$i,@ibuf,@rbuf,$ntot,$vi) = (0,0,0);
  my ($limrb) = $nri*$lenbuf;
  $ntot = @$ix;
  $ibuf[$lenbuf-1] = 0;
  $rbuf[$limrb-1] = 0.0e0;
    for ($i=0; $i < $ntot; $i++) {
      $vi = $ix->[$i];
      if ($vi != 0.0e0) {
	$ibuf[$indib++] = $i+1;
	if ($nri == 1) {$rbuf[$indrb++] = $vi;}
	else {$rbuf[$indrb++] = Re($vi); $rbuf[$indrb++] = Im($vi);}
	$nval++;
	if ($indib == $lenbuf) {
	  if ($dowrt) {$self->_rwvec(1,"i",$lenbuf,\@ibuf); $self->_rwvec(1,"d",$nri*$lenbuf,\@rbuf);}
	  $indib = 0; $indrb=0;}
      }
    }
  if ($dowrt && ($indib >0)) {
    for (; $indib < $lenbuf; $indib++) {$ibuf[$indib] = 0;}
    for (; $indrb < $limrb; $indrb++) {$rbuf[$indrb] = 0.0e0;}
    $self->_rwvec(1,"i",$lenbuf,\@ibuf); $self->_rwvec(1,"d",$nri*$lenbuf,\@rbuf);}
  return ($nval);
}

#
# Read a vector of values (integer or real), rounded up to
# an even number of records, and return a reference to
# the vector.  Arguments:
# $type   -- "i" or "d" for integer or double.
# $n      -- number of values to return.
# $lenbuf -- number of items in a record.  An even multiple
#            of lenbuf values are read from the file.
#
sub _readbufvec {
  my $self = shift;
  my ($nri,$type,$n,$lenbuf) = @_;
  my ($fh,$buf,$i,$j,@x,$ndo) = $$self{"__FH"};
  if ($nri == 2) {$x[$n-1]=cplx(0,0);}
  elsif ($type eq "d") {$x[$n-1]=0.0e0;}
  else {$x[$n-1]=0;}
  for ($i = 0; $i < $n; $i += $lenbuf) {
    $buf = $self->_readvec($type,$nri*$lenbuf);
    $ndo = $lenbuf;
    if (($i+$ndo) > $n) {$ndo = $n - $i;}
    if ($nri == 1) {for ($j = 0; $j < $ndo; $j++) {$x[$i+$j] = $$buf[$j];}}
    else {
      my ($real,$imag,$indji);
      $indji = 0;
      for ($j = 0; $j <$ndo; $j++) {
	$real = $$buf[$indji++];
	$imag = $$buf[$indji++];
	$x[$i+$j] = cplx($real,$imag);
      }
    }
  }
  return (\@x);
}

#
# Read an AO 2e integral block.  These have 4 integer indices and sets
# of 1-3 double values, the latter depending on whether the integrals
# are regular, Rafenetti 1, 2, or 3.  Arguments:
#
# $label  -- string for label record, used to determine the type
#            of the integrals.
# $ni     -- number of integer values from label record; must be 4.
# $nr     -- number of double values from label record, used to
#            determine which Raf. type is present.
# $val    -- number of non-zero integral or integral sets are stored.
# $lenbuf -- number of integral sets in each record.  An even number
#            of records (each with 4*$lenbuf integers and $nr*$lenbuf
#            doubles.
#
# returns a reference to an array with all regular/raff1 integrals
# first, possibly followed by raff2 and raff3 integrals.
#
sub _read2e {
  my $self = shift;
  my ($label,$ni,$nr,$nval,$lenbuf,$n1) = @_;
  my ($fh) = $$self{"__FH"};
  my ($reg) = ($label =~ /REG/) ? "Int" : "R1";
  my ($ival,$ind,$buf,$ndo,$jval,$line,@x,$i,$j,$k,$l,$ij,$kl,$ijkl,$indiv,$ir,$ntot,$ntt);
  my ($nb) = $$self{"__NBASIS"};
  $n1 = abs($n1);
  if ($n1 != $nb) {croak "MatEl::_read2e is confused by n1 $n1 nbasis $nb\n";}
  $ntt = ($n1*($n1+1))/2;
  $ntot = ($ntt*($ntt+1))/2;
  $x[$nr*$ntot-1] = 0.0e0;
  for ($ijkl=0 ; $ijkl<($nr*$ntot); $ijkl++) {$x[$ijkl] = 0.0e0};
  for ($ival = 0; $ival < $nval; $ival += $lenbuf) {
    $ind = $self->_readvec("i",4*$lenbuf);
    $buf = $self->_readvec("d",$nr*$lenbuf);
    $ndo = $lenbuf;
    if (($ndo+$ival) > $nval) {$ndo = $nval - $ival;}
    for ($jval = 0; $jval < $ndo; $jval++) {
      ($i,$j,$k,$l) = ($$ind[4*$jval],$$ind[4*$jval+1],$$ind[4*$jval+2],$$ind[4*$jval+3]);
      $ij = ($i > $j) ? ($i*($i-1))/2 + $j : ($j*($j-1))/2 + $i;
      $kl = ($k > $l) ? ($k*($k-1))/2 + $l : ($l*($l-1))/2 + $k;
      $ijkl = ($ij > $kl) ? ($ij*($ij-1))/2 + $kl : ($kl*($kl-1))/2 + $ij;
      $ijkl = $ijkl - 1;
      $indiv = $nr*$jval;
      for ($ir = 0; $ir < $nr; $ir++) {$x[$ijkl+$ir*$ntot] = $$buf[$indiv++];}
    }
  }
  return (\@x);
}

sub aoint {
  my ($label) = @_;
  my ($val) = (($label =~ /^REG/) || ($label =~ /^RAF/)) ? 1 : 0;
  return ($val);
}

#
# Write a vector of values buffered/block.
#
sub _writebufvec {
  my $self = shift;
  my ($nri,$type,$ntot,$lenbuf,$vals) = @_;
  my ($i,$j,$ndo,$indji,$typew,@buf,$zero,$vj);
  if ($type eq "i") {$zero = 0;}
  else {$zero = 0.0e0;}
  $buf[$nri*$lenbuf-1] = $zero;
  $vj = @$vals;
  if ($ntot != $vj) {croak "Inconsistent ntot $ntot and nvals $vj in MatEl::_writebufvec\n";}
  for ($i = 0; $i < $ntot; $i += $lenbuf) {
    $ndo = $lenbuf;
    if (($i+$ndo) > $ntot) {$ndo = $ntot - $i;}
    $indji = 0;
    if ($nri == 2) {
      $typew = "d";
      for ($j = 0; $j < $ndo; $j++) {
	$vj = $$vals[$i+$j];
	$buf[$indji++] = Re($vj);
	$buf[$indji++] = Im($vj);}}
    else {
      $typew = $type;
      for ($j = 0; $j < $ndo; $j++) {$buf[$indji++] = $$vals[$i+$j];}}
    for (; $indji < ($nri*$lenbuf); $indji++) {$buf[$indji] = $zero;}
    $vj = @buf;
    $self->_rwvec(1,$typew,$nri*$lenbuf,\@buf);
  }
}

sub _write2e {
  my $self = shift;
  my ($dowrt,$nr,$nb,$lenbuf,$vals) = @_;
  my ($indib,$indrb,$ntot,$ijkl,@ibuf,@rbuf,$i,$j,$k,$l,$ntt,$n4,$liml,$v1,$v2,$v3) = (0,0,0,-1);
  my ($limib,$limrb) = (4*$lenbuf,$nr*$lenbuf);
  $ibuf[$limib-1] = 0;
  $rbuf[$limrb-1] = 0.0e0;
  my($thresh) = $self->INTTHRESH;
  if (! defined($thresh)) {$thresh = 1.0e-12;}
  $ntt = ($nb*($nb+1))/2;
  $n4 = ($ntt*($ntt+1))/2;
  $v2 = 0.0e0;
  $v3 = 0.0e0;
  for ($i=1; $i <= $nb; $i++) {
    for ($j=1; $j <= $i; $j++) {
      for ($k=1; $k <= $i; $k++) {
	$liml = ($k == $i) ? $j : $k;
	for ($l=1; $l <= $liml; $l++) {
	  $ijkl++;
	  $v1 = $vals->[$ijkl];
	  if ($nr > 1) {$v2 = $vals->[$ijkl+$n4];}
	  if ($nr == 3) {$v3 = $vals->[$ijkl+2*$n4];}
	  if ((abs($v1) > $thresh) || (abs($v2) > $thresh) || (abs($v3) > $thresh)) {
	    $ibuf[$indib++] = $i;
	    $ibuf[$indib++] = $j;
	    $ibuf[$indib++] = $k;
	    $ibuf[$indib++] = $l;
	    $rbuf[$indrb++] = $v1;
	    if ($nr > 1) {$rbuf[$indrb++] = $v2;}
	    if ($nr == 3) {$rbuf[$indrb++] = $v3;}
	    $ntot++;
	    if ($indib == $limib) {
	      if ($dowrt) {$self->_rwvec(1,"i",$limib,\@ibuf); $self->_rwvec(1,"d",$limrb,\@rbuf);}
	      $indib = 0;
	      $indrb = 0;}
	  }
	}
      }
    }
  }
  if ($dowrt && ($indib > 0)) {
    for (; $indib < $limib; $indib++) {$ibuf[$indib] = 0;}
    for (; $indrb < $limrb; $indrb++) {$rbuf[$indrb] = 0.0e0;}
    $self->_rwvec(1,"i",$limib,\@ibuf); $self->_rwvec(1,"d",$limrb,\@rbuf);}
  return ($ntot);
}

sub _intname {
  my ($name,$num) = @_;
  $name = "__" . uc($name);
  if (defined($num)) {$name .= "$num";}
  return ($name);
}

sub header {
  my $self = shift;
  my ($item,$oarr,$nlab,$namerec);
  if (@_) {
    foreach $item (@head_scalars) {$$self{_intname($item)} = shift;}
    foreach $item (@head_arrays) {$oarr = shift; my (@arr) = @$oarr; $$self{_intname($item)} = \@arr;}
    if (@_) {
      $nlab = 9;
      $nlab++;
      foreach $item (@_) {$nlab++; my (@arr) = @$item; $namerec = _intname("REC".$nlab); $$self{$namerec} = \@arr;}
      $$self{"__NLAB"} = $nlab;
    }
  }
  my (@res);
  foreach  $item (@head_scalars,@head_arrays) {push(@res,$$self{_intname($item)});}
  $nlab = $$self{"__NLAB"};
  for ($item=11; $item <= $nlab; $item++) {
    push(@res,$$self{_intname("rec",$item)});
  }
  return (@res);
}

sub copy_header {
  my $self = shift;
  my ($old) = @_;
  $self->header($old->header());
}

sub item {
  my $self = shift;
  my $type = ref($self) or croak "$self is not an object";
  my $name = shift;
  $name = uc($name);
  my ($names) = $name;
  $names =~ s/ //g;
  my ($arg) = @_;
  if (defined($scalars{$names})) {
    if (! defined($self->{$GSNAMEX})) {
      my (@a,$i);
      $a[999] = 0.0e0;
      for ($i=0; $i < 999; $i++) {$a[$i]=0.0e0;}
      $self->{$GSNAMEX} = \@a;
    }
    my ($s) = $self->{$GSNAMEX};
    if (defined($arg)) {return $s->[$scalars{$names}] = $arg;}
    else {return $s->[$scalars{$names}];}
  }
  else {
    $name = "_" . $name;
    my ($name1) = "_" . $name;
    if (defined($arg)) {
      if (defined($self->{$name1})) {return $self->{$name1} = $arg;}
      else {return $self->{$name} = $arg;}}
    elsif (defined($self->{$name1})) {return $self->{$name1};}
    elsif (defined($self->{$name})) {return $self->{$name};}
    else {return undef;}
  }
}

sub AUTOLOAD {
  my $self = shift;
  my $name = $AUTOLOAD;
  my $type = ref($self) or croak "$self is not an object in $name";
  $name =~ s/.*://;   # strip fully-qualified portion
  return($self->item($name,@_));
}

# build arrays of arrays holding connectivty -- a list of bonded atom
# bond type pairs for each atom.
sub _makecon {
  my $self = shift;
  my ($natoms,$nbtot,$nbondo,$ibondo,$rbtypeo,$indi,$indr) =
    ($self->natoms,$self->nbondtot,$self->item("BONDS PER ATOM"),
    $self->item("BONDED ATOMS"),$self->item("BOND TYPES"),0,0);
  my ($nbond,$ibond,$rbtype) = ($nbondo->array(),$ibondo->array(),$rbtypeo->array());
  my ($i,$j,$nbtoti,$nbtotr,@atcon);
  if (($nbtot != @$ibond) || ($nbtot != @$rbtype)) {die "mismatch nbtot $nbtot $nbtoti $nbtotr\n";}
  $atcon[$natoms-1] = undef;
  for ($i=0; $i < $natoms; $i++) {
    my (@atl);
    for ($j=0; $j < $nbond->[$i]; $j++) {
      my (@pair) = ($ibond->[$indi++],$rbtype->[$indr++]);
      push (@atl,\@pair);}
    $atcon[$i] = \@atl;}
  $self->{_intname("connectivity")} = \@atcon;
}

#
# Print the connectivity item.
#
sub prtcon {
  my $self = shift;
  if (defined($self->{_intname("connectivity")})) {printcon ($self->natoms,$self->connectivity);}
  else {print "No connectivity available\n";}
}

# Print a connectivity
sub printcon {
  my ($natoms,$con) = @_;
  my ($i,$pr);
  for ($i=0; $i<$natoms; $i++) {
    print "connectivity for atom ",$i+1,":";
    foreach $pr (@{$con->[$i]}) {printf("%7d %5.2f",$$pr[0],$$pr[1]);}
    print "\n";}
}

1;
