#
# Class OpMat and utility routines for individual or related operator
# matrices in matrix element file.
#
# Methods:
#
# new($name,$type,$dimens,$values) -- create object:
#     $name   -- string to name object
#     $type   -- "i", "d", or "c" for integer, double
#                or complex.
#     $dimens -- ref for array of dimensions
#     $values -- ref for array of values
#
# dimens -- return/set array of dimensions.  If setting,
#           total length must be unchanged.
#
# get_elem($i,$j,...) -- fetch element
#
# set_elem(value,$i,$j,...) -- set element
#
# make_indx($i,$j,...) -- return index into values array
#
# lenarray() -- return number of values stored
#
# array -- return reference to entire array of values
#
# name -- return name of object
#
# type -- return type of contents:  i, d, or c
#         for integer, double, or complex
#
# Importable functions:
#
# lind2(i,j)     -- indexing for lower-triangular storage:
#                   i,j start at 1, output is for 0 starting index.
#
# lind3(i,j,k)   -- indexing for lower-tetrahedral storage:
#                   i,j,k start at 1, output is for 0 starting index.
#
# lind4(i,j,k,l) -- indexing for lower-4dim storage, totally symmetric:
#                   i,j,k,l start at 1, output is for 0 starting index.

package OpMat;
use strict;
use warnings;
use Carp;
use Exporter;

our ($AUTOLOAD,@EXPORT_OK,@ISA);

@ISA = qw(Exporter);

@EXPORT_OK = qw(&lind2 &lind3 &lind4 &fixstr &nlt);

sub new {
    my $this = shift;
    my $class = ref($this) || $this;
    my $self = {};
    bless $self, $class;
    $self->_init();
    if (@_) {$self->_load(@_);}
    return $self;
  }

sub _init {
  my $self = shift;
  my $key;
  foreach $key (keys %$self) {delete($$self{$key});}
  $$self{"_NAME"} = undef;
  return ($self);
}

sub _load {
  my $self = shift;
  my ($name,$type,$dimens,$values) = @_;
  if (! $values) {croak "Bad arguments to OpMat::_load\n";}
  $$self{"_NAME"} = $name;
  if (($type ne "i") && ($type ne "d") && ($type ne "c")) {croak "Bad type $type to OpMat::_load\n";}
  $$self{"_TYPE"} = $type;
  my (@dim,$x);
  if (ref($dimens)) {foreach $x (@$dimens) {if ($x ne 0) {push(@dim,$x);}}}
  else {push(@dim,$dimens);}
  $$self{"_DIMENS"} = \@dim;
  my ($ltot,$lval) = $self->lenarray();
  $lval = @$values;
  if ($ltot != $lval) {croak "ltot=$ltot not lval=$lval for $name\n";}
  $$self{"_VALUES"} = $values;
  return ($self);
}

sub lenarray {
  my $self = shift;
  my ($iflt) = 1;
  my ($ltot,$dim,$ntt,$i,$last) = (0,$$self{"_DIMENS"},1);
  my ($nlt) = nlt($iflt,$dim);
  if ($nlt == -2) {$ntt = (lind2($$dim[1],$$dim[1])+1)*(lind2($$dim[3],$$dim[3])+1); $nlt=3;}
  elsif ($nlt ge 0) {
    if ($nlt == @$dim) {croak "Bad dimens in OpMat::lenarray\n";}
    elsif ($nlt == 0) {$ntt = lind2($$dim[1],$$dim[1]) + 1; $nlt++;}
    elsif ($nlt == 1) {$ntt = lind3($$dim[2],$$dim[2],$$dim[2]) + 1; $nlt++;}
    elsif ($nlt == 2) {$ntt = lind4($$dim[3],$$dim[3],$$dim[3],$$dim[3]) + 1; $nlt++;}
    elsif ($nlt > 0) {croak "OpMat::lenarray does not understand nlt=$nlt\n";  $nlt++;}
  }
  if ($nlt != -2) {for ($i=$nlt+1; $i < @$dim; $i++) {$ntt *= abs($$dim[$i]);}}
  return ($ntt);
}

# return index of last lt index in dimensions, or -2 for -+-+
sub nlt {
  my ($iflt,$dim) = @_;
  my ($nlt,$i) = (-1);
  if ($iflt) {
    if ((@$dim == 5) && ($$dim[0] < 0) && ($$dim[1] > 0) && ($$dim[2] < 0) && ($$dim[3] > 0)) {$nlt = -2;}
    else {      
      for ($i=0; $i < @$dim; $i++) {
        if ($$dim[$i] < 0) {$nlt++;}
        else {last;}
      }
    }
    if ($nlt != -2) {
      for ($i=($nlt+1); $i < @$dim; $i++) {
        if ($$dim[$i] < 0) {croak "bad dims in OpMat::nlt ",join(" ",@$dim),"\n";}}}
  }
  return ($nlt);
}

sub array {my $self = shift; return($$self{"_VALUES"});}

sub get_elem {
  my $self = shift;
  my ($val,$indx) = ($$self{"_VALUES"});
  $indx = $self->make_indx(@_);
  return ($val->[$indx]);
}

sub set_elem {
  my $self = shift;
  my $nval = shift;
  my ($val,$indx) = ($$self{"_VALUES"});
  $indx = $self->make_indx(@_);
  return ($val->[$indx]=$nval);
}

sub make_indx {
  my ($self) = shift;
  my (@inds) = @_;
  my ($dim,$name) = ($self->{"_DIMENS"});
  if (@inds != @$dim) {$name = $self->NAME(); croak "dimension error for $name\n";}
  my ($prod,$indx,$i,$d,$nd,$nlt) = (1,0);
  $nlt = nlt(1,$dim);
  $nd = @inds;
  if ($nlt >= ($nd-1)) {croak "Bad dimens in OpMat::make_indx\n";}
  $indx = 0;
# if ($nlt == -2) {
#   $ind1 = lind2(
#   }
  if ($nlt < ($nd-2)) {
    $indx = $inds[$nd-1] - 1;
    for ($i = $nd-2; $i > ($nlt+1); $i--) {$indx *= $dim->[$i+1]; $indx = $indx + $inds[$i] - 1;}}
  if ($nlt ge 0) {
    if ($nlt == @$dim) {croak "Bad dimens in OpMat::make_indx\n";}
    if ($nlt == 0) {
      $indx = $indx*(lind2($$dim[1],$$dim[1])+1) + lind2($inds[0],$inds[1]);}
    elsif ($nlt == 1) {
      $indx = $indx*(lind3($$dim[2],$$dim[2],$$dim[2])+1) + lind3($inds[0],$inds[1],$inds[2]);}
    elsif ($nlt == 2) {
      $indx = $indx*(lind4($$dim[3],$$dim[3],$$dim[3],$$dim[3])+1) + lind4($inds[0],$inds[1],$inds[2],$inds[3]);}
    else {croak "OpMat::make_indx does not understand nlt=$nlt\n";}
  }
  return ($indx);
}

sub lind2 {my ($i,$j) = @_; if ($i < $j) {($j,$i) = @_;} return ($j-1+($i*($i-1))/2);}

sub lind3 {
  my ($ii,$ji,$ki,$sum,$i,$j,$k) = @_;
  $k = ($ki > $ji) ? $ki : $ji; $k = ($k > $ii) ? $k : $ii;
  $i = ($ki > $ji) ? $ji : $ki; $i = ($i > $ii) ? $ii : $i;
  $j = $ii + $ji + $ki - $k - $i;
  return($i-1+($j*($j-1))/2+(($k-1)*$k*($k+1))/6);
}

sub lind4 {my ($i,$j,$k,$l) = @_; return (lind2(lind2($i,$j)+1,lind2($k,$l)+1));}

sub lablist {
  my $self = shift;
  my ($lenbuf) = @_;
  my (@dims) = (0,0,0,0,0,0);
  my ($nri,$vals,$oname,$ni,$nr,$ntot,$lenbufu,$otype,$i,$odims);
  $oname = fixstr(64,$self->name());
  $otype = $self->type();
  $odims = $self->dimens();
  $vals = $self->array();
  $ni = 0;
  $lenbufu = $lenbuf;
  if ($otype eq "c") {$ni = 0; $nr = -1; $nri = 2; $lenbufu = $lenbuf/4;}
  elsif ($otype eq "d") {$ni = 0; $nr = 1; $nri = 1; $lenbufu = $lenbuf/2;}
  else {$ni = 1; $nr = 0; $nri = 1;}
  $ntot = @$vals;
  for ($i=0; $i < @$odims; $i++) {$dims[$i] = $odims->[$i];}
  return ($nri,$vals,$oname,$ni,$nr,$ntot,$lenbufu,@dims);
}

sub fixstr {
  my ($len,$instr,$outstr,$fmt) = @_;
  $fmt = "%-" . $len . "s";
  if (! defined($instr)) {$instr = " ";}
  $outstr = sprintf($fmt,$instr);
  return (substr($outstr,0,$len));
}

sub AUTOLOAD {
  my $self = shift;
  my $type = ref($self) or croak "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://;   # strip fully-qualified portion
  $name = uc($name);
  $name = "_" . uc($name);
  if (@_) {return $self->{$name} = shift;}
  elsif (defined($self->{$name})) {return $self->{$name};}
  else {return undef;}
}

1;
