#!/usr/bin/env python3
#
# This is an example of use of the Python interface to Gaussian.
# It either picks up a matrix element file from a previous job
# or runs job on a simple molecule (HOF).  In either case, it picks
# up the MO 2e integrals and orbital energies and computes the
# MP2 energy.
#
# Note that this script and the underlying modules require
# python3.
#
import sys
import numpy as np
import qcmatrixio as qcio
import QCOpMat as qco
import QCMatEl as qcm

def main():
# Setting debug True will cause the intermediate matrix element
# and Gaussian input and output files to be named and saved.
  debug = False
  savefiles = debug

# The name of a matrix element file computed by Gaussian in a
# separate calculation can be specified on the command line.
  if len(sys.argv) >= 2:
    fname = sys.argv[1]
#   Create the container object with the contents of the
#   specified file
    mel = qcm.MatEl(file=fname)

  else:
#   Set up a container object with the minimal information
#   specifying a molecule -- coordinates, atomic numbers, etc.
#   and then run Gaussian to populate the object with results.

#   When debugging, we want to save intermediate files.  By
#   default, temporary files are generated and discarded.
    if savefiles:
      fnamei = "domp2i.mat"
      fname = "domp2o.mat"
      inpname = "test.inp"
      outname = "test.out"
    else:
      fnamei = None
      fname = None
      inpname = None
      outname = None

#   Create an empty matrix element object
    mel = qcm.MatEl()

#   Set the minimal information to do a calculation on HOF
#   radical cation in local variables.  Quantities which
#   are not scalars are stored in flat (1-dimensional)
#   numpy arrays.
    nat = 3     # number of atoms
    iopcl = 1   # open-shell (uhf) calculation
    mul = 2     # multiplicity 2
#   atomic numbers
    iana = np.array([8,9,1],dtype="int32")
#   coordinates in Angstroms, converted to Bohr
    cang = np.array([0.0,0.0,0.0,0.0,0.0,1.42,0.905097,0.0,-0.32],dtype="float64")
    atcora = np.array([cx / 0.52917721092 for cx in cang],dtype="float64")
#   number of electrons, adjusted for the net charge required to
#   make an odd number of electrons
    nelec = sum(ia for ia in iana)
    if ((nelec % 2) == 0): ich = 1
    else: ich = 0
    nelec -= mel.icharg

#   Put the minimal data into the object
    mel.load_head(natoms=nat,icharg=ich,multip=mul,ne=nelec,ian=iana,
                  iopcl=iopcl,c=atcora)

#   Do the calculation and update the data in the object
#   Except for debugging, the matfi, matfo, tinput, and
#   toutput arguments could be omitted as they default to None
#   mel is automatically updated with all results left by the
#   calculation.
#   The commented-out line illustrates the inclusion of optional
#   command-line arguments to Gaussian.
    mel.update(matfi=fnamei,matfo=fname,doinit=True,motran="full",
               basis="3-21G",dofock="SCF",tinput=inpname,
               progargs=["-c=2-3","-m=64mw"],
               toutput=outname)

# Both the container object and the objects for each item
# within it have appropriate print methods and can be
# printed in the usual way.
  if debug:  print("matrix element object",mel)

# compute derived scalars from the values in the object
  nae = int(int(mel.ne+mel.multip-1)/2)  # number of alpha electrons
  nbe = int(int(mel.ne-mel.multip+1)/2)  # number of beta electrons
  noa = nae - mel.nfc                    # number of active (not frozen
  nob = nbe - mel.nfc                    # core or virtual) electrons
  nrorb = mel.nbsuse - mel.nfc - mel.nfv # number of active orbitals
  nva = nrorb - noa                      # number of active alpha virtuals
  nvb = nrorb - nob                      # number of active betavirtuals
  print ("NAtoms=",mel.natoms,"NEl=",mel.ne,"NAE=",nae,"NBE=",nbe,
         "NFC=",mel.nfc,"NFV=",mel.nfv,"NOA=",noa,"NOB=",nob,"NVA=",nva,
         "NVB=",nvb)

# Recover the 5 arrays we need, expanded to full index ranges
# (not lower triangles) and dimensioned for easy access.
# mel.itran indicates whether a full or partial transformation
# was performed

  if mel.itran == 5:
    ldima = nrorb
    ldimb = nrorb
  else:
    ldima = noa
    ldimb = nob

  aoe = mel.matlist["ALPHA ORBITAL ENERGIES"].expand().reshape((mel.nbsuse))
  aa2e = mel.matlist["AA MO 2E INTEGRALS"].expand().reshape((ldima,nrorb,nrorb,nrorb))
  if mel.iopcl == 1:
    ba2e = mel.matlist["BA MO 2E INTEGRALS"].expand().reshape((ldima,nrorb,nrorb,nrorb))
    boe = mel.matlist["BETA ORBITAL ENERGIES"].expand().reshape((mel.nbsuse))
    bb2e = mel.matlist["BB MO 2E INTEGRALS"].expand().reshape((ldimb,nrorb,nrorb,nrorb))

# Compute and print the mp2 energy and t2 contribution for each spin-case
  te2aa = domp2_xx (mel.nfc,noa,nva,aoe,aa2e)
  if mel.iopcl == 1:
    te2ab = domp2_xy (mel.nfc,noa,nob,nva,nvb,aoe,boe,ba2e)
    te2bb = domp2_xx (mel.nfc,nob,nvb,boe,bb2e)
  else:
    te2ab = domp2_xy (mel.nfc,noa,nob,nva,nvb,aoe,aoe,aa2e)
    te2bb = te2aa

  print ("t2aa",te2aa[0],"e2aa",te2aa[1])
  print ("t2ab",te2ab[0],"e2ab",te2ab[1])
  print ("t2bb",te2bb[0],"e2bb",te2bb[1])

# Recover the scf energy from the Gaussian calculation, report the
# total energy, and check the mp2 energy if there was a value
# in the file
  ehf = mel.scalar("escf")
  emp2 = ehf + te2aa[1] + te2ab[1] + te2bb[1]
  eref = mel.scalar("eump2")
  if (eref == 0.0): print ("EHF",ehf,"EMP2",emp2)
  else: print ("EHF",ehf,"EMP2",emp2,"Err",emp2-eref)

# Compute the contribution to T2 and E2 from a pure-spin case:
def domp2_xx (nfc,no,nv,oe,x2e):
  e2 = 0.0
  t2 = 0.0
  for i in range (1,no):
    for j in range (0,i):
      eij = oe[i+nfc] + oe[j+nfc]
      for a in range (1,nv):
        aa = a + no
        eija = oe[a+nfc+no] - eij
        for b in range (0,a):
          bb = b + no
          eijab = oe[b+nfc+no] + eija
          tei = x2e[j,bb,i,aa] - x2e[j,aa,i,bb]
          e2 = e2 - tei*tei/eijab
          t2 = t2 + tei*tei/(eijab*eijab)
  return (t2,e2)

# Compute the contribution to T2 and E2 from the mixed -spin case:
def domp2_xy (nfc,noa,nob,nva,nvb,aoe,boe,x2e):
  e2 = 0.0
  t2 = 0.0
  for i in range (0,noa):
    for j in range (0,nob):
      eij = aoe[i+nfc] + boe[j+nfc]
      for a in range (0,nva):
        aa = a + noa
        eija = aoe[a+nfc+noa] - eij
        for b in range (0,nvb):
          bb = b + nob
          eijab = boe[b+nfc+nob] + eija
          tei = x2e[i,aa,j,bb]
          e2 = e2 - tei*tei/eijab
          t2 = t2 + tei*tei/(eijab*eijab)
  return (t2,e2)

main()
