#
# This file contains an example of using the python interface
# to Gaussian and the matrix element file.  It provides
# routines which can be used with the PyQuante2 program to
# compute one and two electron integrals using Gaussian
# instead of the integral code provided with PyQuante2.
# The routines provided are:
#
# pq2_to_gau (geo,bfs,debug=False,doupdate=True,**kwargs)
#
# this takes pyquante2 objects for the molecule and basis
# set and returns a MatEl object.  If doupdate is true,
# then the initial object is run through Gaussian and the
# returned object includes the results.  Keyword arguments
# are passed on to the update procedure and determine what
# is calculated.
#
# and:
#
# pq2_onee_integrals (mel)
# pq2_twoe_integrals (mel)
#
# these take a matrix element object and return 1e or 2e
# integrals in pyquante2 format.
#
import sys
import subprocess
import numpy as np
import QCOpMat as qco
import QCMatEl as qcm
import datetime

# gaussian order of cartesian bfns for pyquante2 interface:
pq2_pow2num = {(0,0,0):    1, (1,0,0): 1001, (0,1,0): 1002,
               (0,0,1): 1003, (2,0,0): 2001, (0,2,0): 2002,
               (0,0,2): 2003, (1,1,0): 2004, (1,0,1): 2005,
               (0,1,1): 2006, (3,0,0): 3001, (0,3,0): 3002,
               (0,0,3): 3003, (1,2,0): 3004, (2,1,0): 3005,
               (2,0,1): 3006, (1,0,2): 3007, (0,1,2): 3008,
               (0,2,1): 3009, (1,1,1): 3010}

def pq2_to_gau (geo,bfs,debug=False,doupdate=True,**kwargs):
  mel = qcm.MatEl(debug=debug)
  natoms = len(geo.atoms)
  nel = geo.nel()
  iopcl = nel % 2
  ian = [atom.atno for atom in geo.atoms]
  c = np.array([atom.r for atom in geo.atoms],dtype="float64").reshape((3*natoms))
  nbasis = len(bfs.bfs)
  mel.load_head (title=geo.name,natoms=natoms,icharg=geo.charge,nbasis=nbasis,
                 multip=geo.multiplicity,ne=nel,iopcl=iopcl,ian=ian,c=c,
                 labfil="From PyQuante2")
  mel.nshlao = len(bfs.shells)
  mel.nprmao = sum(len(sh.coefs) for sh in bfs.shells)
  mel.ibfatm = np.array([pq2_r_to_at(bf.origin,geo) for bf in bfs.bfs],dtype="int32")
  mel.ibftyp = np.array([pq2_pow2num[bf.powers] for bf in bfs.bfs],dtype="int32")
  mel.addobj(qco.OpMat("SHELL TO ATOM MAP",np.array([pq2_r_to_at(shl.origin,geo) for shl in bfs.shells],dtype="int32")))
  mel.addobj(qco.OpMat("SHELL TYPES",np.array([shl.am for shl in bfs.shells],dtype="int32")))
  mel.addobj(qco.OpMat("NUMBER OF PRIMITIVES PER SHELL",np.array([len(shl.exps) for shl in bfs.shells],dtype="int32")))
  exps = []
  ccs = []
  shlxyz = []
  for shl in bfs.shells:
    for exp in shl.exps:  exps.append(exp)
    for cc in shl.coefs:  ccs.append(cc)
    for x in shl.origin:  shlxyz.append(x)
  mel.addobj(qco.OpMat("PRIMITIVE EXPONENTS",np.array(exps,dtype="float64")))
  mel.addobj(qco.OpMat("CONTRACTION COEFFICIENTS",np.array(ccs,dtype="float64")))
  mel.addobj(qco.OpMat("P(S=P) CONTRACTION COEFFICIENTS",np.zeros((mel.nprmao),dtype="float64")))
  mel.addobj(qco.OpMat("COORDINATES OF EACH SHELL",np.array(shlxyz,dtype="float64")))
  mel.set_scalars(enucrep=geo.nuclear_repulsion())
  if debug: print ("input matrix element file:",mel)
  if doupdate:
    mel.update(**kwargs)
    if debug: print ("output matrix element file:",mel)
  return (mel)

def pq2_r_to_at (r,geo):
# match up shell/basis function centers to atomic coordinates.
  for i,at in enumerate (geo.atoms):
    if at[0] == r[0] and at[1] == r[1] and at[2] == r[2]: return(i+1)
  assert False

# return an object compatible with that returned by onee_integrals
# with pyquante2.
class pq2_onee_integrals (object):
  def __init__(self,mel):
    nbf = self.nbf = mel.nbasis
    self.S = mel.matlist["OVERLAP"].expand().reshape((nbf,nbf))
    self.T = mel.matlist["KINETIC ENERGY"].expand().reshape((nbf,nbf))
    self.V = mel.matlist["CORE HAMILTONIAN ALPHA"].expand().reshape((nbf,nbf))
    self.V = self.V - self.T

# return an object compatible with that returned by twoe_integrals
# with pyquante2.
class pq2_twoe_integrals (object):
  def __init__(self,mel,timer):
    nbf = self.nbf = mel.nbasis
    self._2e_ints = mel.matlist["REGULAR 2E INTEGRALS"].expand().reshape((nbf,nbf,nbf,nbf))
#   timer.repdelta ("mel2e")
#   mel2e.expand()
#   timer.repdelta ("expand")
#   self._2e_ints = mel2e.array.reshape((nbf,nbf,nbf,nbf))
#   timer.repdelta ("reshape")
    return

  def __getitem__(self,*args): return self._2e_ints.__getitem__(*args)
  def __repr__(self): return repr(self._2e_ints.ravel())

  def transform(self,c): return np.einsum('aI,bJ,cK,dL,abcd->IJKL',c,c,c,c,self._2e_ints)
  def transform_mp2(self,c,nocc):
    return np.einsum('aI,bJ,cK,dL,abcd->IJKL',c[:,:nocc],c,c[:,:nocc],c,self._2e_ints)

  def get_j(self,D): return np.einsum('kl,ijkl->ij',D,self._2e_ints)
  def get_k(self,D): return np.einsum('ij,ikjl->kl',D,self._2e_ints)
  def get_2jk(self,D): return 2*self.get_j(D)-self.get_k(D)

class qctimer:
  def __init__ (self):
    self.oldtime = datetime.datetime.today()

  @property
  def curdelta (self):
    return ((datetime.datetime.today()-self.oldtime).total_seconds())

  @property
  def delta (self):
    now = datetime.datetime.today()
    delta = now - self.oldtime
    oldtime = now
    return (delta.total_seconds())

  def repdelta (self,str,widthstr=10,widthfp=4,**kwargs):
    fstr = "%%-%d%s %%%d.2f sec" % (widthstr,"s",widthfp)
    print (fstr % (str+":",self.delta),**kwargs)
